/*
   Shadow Volumes Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to create shadows using a technique called stencil
   shadow volumes in a 3d scene.  To do this we will not need the shadow matrix from
   previous demo.  In fact in order to perform this all we need is to render the
   silhouette edges of an object (which are the outer edges of a object) extended in
   a direction (which depends on the light position) into the stencil buffer twice.  After
   that we draw the scene twice, once for the dark parts of the scene (areas that are 
   shadowed) and once for the lit parts of the scene (areas that are not in a shadow).
   With that you can create extremely realistic shadows in your scenes.  At first glance
   this sounds a lot more complicated than it really is.  When it comes to the silhouette
   edges we don't need to worry about that because the object casting a shadow in this
   demo is a simple quad.  So its outer edges are the same as every point that makes
   it up.

   Controls:

      Up Arrow    - move the light source back.
      Down Arrow  - move the light source forward.
      Left Arrow  - move the light source left.
      Right Arrow - move the light source right.
      Space Bar   - Render the Shadow Volume.
      Escape      - exit the application.
*/


// Demo: Shadow Volumes.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_EXTRALEAN       // Trims even further.


#include<windows.h>        // Windows header file.
#include<gl/gl.h>          // OpenGL header.
#include<gl/glu.h>         // OpenGL utility.
#include<gl/glaux.h>       // OpenGL auxiluary.
#include"CVector.h"        // Vector class.
#include"CTGALoader.h"     // Texture object CTexture.


// Function prototypes...
bool InitializeGL();
void RenderScene();
void Render();
void DrawSilhouette(CVector4 lightPos, float Extend);
void Shutdown();


// The global device context.
HDC g_HDC;


// xRotation and yRotation is used to rotate the entire scene.
float xRotation = 0.0f;
float yRotation = 0.0f;


// The object casting the shadow in this scene is a simple quad.  So this define
// will just hold the max amount of points in a quad...which is 4.
#define MAX_SHADOW_CASTER_POINTS 4

// The position of the main light source
CVector4 lightPosition(2.0f, 5.0f, 0.0f, 1.0f);

// displayVolume is used to visually see the shadow volume.  ExtendAmount is used to
// help create the silhouette of the ojbect that will be casting the shadow.
bool displayVolume = false;
float ExtendAmount = 5.0f;

// This is the object that will be casting the shadow.
CVector4 ShadowObject[MAX_SHADOW_CASTER_POINTS];

// Texture images.
CTGALoader Terrain;
CTGALoader ShadowObjectTex;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // 32 bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         16,                              // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;             // Rendering context.
   static HDC hDC;               // Device context.
   int width, height;            // The window width and height.
	static POINT oldMousePos;     // Last mouse position.
   static POINT currentMousePos; // Current mouse position.
   static bool  isMouseActive;   // Is the left mouse button down.


   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;

						case VK_SPACE:
                    displayVolume = !displayVolume;
                    break;

                  case VK_UP:
                     lightPosition.z -= 0.05f;
                     break;

                  case VK_DOWN:
                     lightPosition.z += 0.05f;
                     break;

                  case VK_LEFT:
                     lightPosition.x -= 0.05f;
                     break;

                  case VK_RIGHT:
                     lightPosition.x += 0.05f;
                     break;


                  case WM_CLOSE:
                  case WM_DESTROY:
                     PostQuitMessage(0);
                     break;

                  default:
                     break;
               }
            break;


			case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
				break;


			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
						xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;


         default: // Always have a default just in case.
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                // A message variable.
   WNDCLASSEX windowClass; // Your Window class.
   HWND hwnd;              // The Window handle.
   bool isFinished;        // This will be used to check if the program is done or not.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create window...
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Shadows Volume Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640, 480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;


   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }

   // Shutdown (free) all resources we used up.
   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.5f, 0.5f, 1.0f, 1.0f);

   glShadeModel(GL_SMOOTH);
   glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);

   // Setup the light source below.
   float ambientLight[] = {1.0f, 1.0f, 1.0f, 1.0f};
   float diffuseLight[] = {1.0, 1.0, 1.0, 1.0}; 
   float specularLight[] = {1.0, 1.0, 1.0, 1.0};

   glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
   glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
   glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);
   glEnable(GL_LIGHT0);


   if(!Terrain.LoadTGA("ground.tga")) return false;
   glGenTextures(1, &Terrain.ID);
   glBindTexture(GL_TEXTURE_2D, Terrain.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   int type = Terrain.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, Terrain.imageWidth, Terrain.imageHeight,
                     type, GL_UNSIGNED_BYTE, Terrain.image);

   if(!ShadowObjectTex.LoadTGA("decal.tga")) return false;
   glGenTextures(1, &ShadowObjectTex.ID);
   glBindTexture(GL_TEXTURE_2D, ShadowObjectTex.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   type = ShadowObjectTex.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, ShadowObjectTex.imageWidth,
                     ShadowObjectTex.imageHeight,
                     type, GL_UNSIGNED_BYTE, ShadowObjectTex.image);

   // This is were we set all the necessary data for the quad that will be
   // casting a shadow on the scene.
	ShadowObject[0] = CVector4(-1.0f, 2.5f, -1.0f);
   ShadowObject[1] = CVector4(-1.0f, 2.5f,  1.0f);
   ShadowObject[2] = CVector4(1.0f, 2.5f,  1.0f);
   ShadowObject[3] = CVector4(1.0f, 2.5f, -1.0f);

   // If all went well we return true.
   return true;
}


void RenderScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glLoadIdentity();

	// Just like any other time we render to the stencil buffer
	// we must first turn off color mask then render the scene.
   glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
   Render();

   // Setup the stencil buffer options.
   glEnable(GL_CULL_FACE);
   glEnable(GL_STENCIL_TEST);
   glDepthMask(GL_FALSE);
   glStencilFunc(GL_ALWAYS, 0, 0);

   // Now here is the key.  We render the shadow volume and increment the stencil
   // where ever a front facing surface is encountered.  The shadow volume is the
   // silhouette of the object extended based on the light position and the amount to extend.
   // So where ever that silhouette is we are taking note of it and incrementing those areas.
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
	glCullFace(GL_BACK);
	DrawSilhouette(lightPosition, ExtendAmount);

	// Now we render the shadow volume again and this time we decrement the stencil every
   // where that is not in the area being rendered.  In other words this is for every
   // thing outside the shadow area.
	glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
	glCullFace(GL_FRONT);
	DrawSilhouette(lightPosition, ExtendAmount);

	// Now we reset everything back to the way it was before we started our stencil test.
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LEQUAL);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	glCullFace(GL_BACK);
	glDisable(GL_CULL_FACE);

	// Here we are rendering the dark parts of the scene.  In other words we are drawing
   // everthing in the shadow volume dark.
   glDisable(GL_LIGHT0);
	glStencilFunc(GL_EQUAL, 1, 1);
	Render();

	// Finally we draw the light parts of the scene.  Unlike before we are now drawing
   // the parts of the scene that is not in the shadow volume (area).
   glEnable(GL_LIGHT0);
	glStencilFunc(GL_EQUAL, 0, 1);
	Render();

	// Since we are done lets turn off the depthfunc and stencil testing.
	glDepthFunc(GL_LESS);
	glDisable(GL_STENCIL_TEST);

   // This is optional.  If you want to see the shadow volume this is true, else false.
	if(displayVolume == true)
		{
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			DrawSilhouette(lightPosition, ExtendAmount);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		}

	SwapBuffers(g_HDC);                 // Display the new frame.
}


void Render()
{
   // This function will draw every object, ground, and light in the scene.
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

   // Apply all translations and rotations.
	glTranslatef(0.0f, -2.0f, -10.0f);
	glRotatef(-yRotation, 1.0f, 0.0f, 0.0f);
	glRotatef(-xRotation, 0.0f, 1.0f, 0.0f);

	// Render the floor...
	glDisable(GL_LIGHTING);
	glPushMatrix();

	   glBindTexture(GL_TEXTURE_2D, Terrain.ID);

      // Draw one big square.
      glBegin(GL_QUADS);
		   glNormal3f(0.0f, 1.0f, 0.0f);
         glTexCoord2f(0.0, 0.0); glVertex3f(15.0f, -0.01f, 15.0f);
         glTexCoord2f(1.0, 0.0); glVertex3f(-15.0f, -0.01f, 15.0f);
         glTexCoord2f(1.0, 1.0); glVertex3f(-15.0f, -0.01f, -15.0f);
         glTexCoord2f(0.0, 1.0); glVertex3f(15.0f, -0.01f, -15.0f);
      glEnd();

	glPopMatrix();
	glEnable(GL_LIGHTING);
	glBindTexture(GL_TEXTURE_2D, NULL);


	// From here on out we will draw the rest of the objects starting with the torus.
   // Note that these objects are not set up to cast shadows of their own but to be
   // shadowed upon.
	glPushMatrix();

		glTranslatef(-2.0f, 0.8f, -0.2f);
		glColor3f(1.0f, 1.0f ,1.0f);
		auxSolidTorus(0.2f, 0.4f);

	glPopMatrix();

   // Next draw the cube next to the torus.
	glPushMatrix();

		glTranslatef(0.0f, 0.8f, -0.2f );
		glColor3f(1.0f, 1.0f ,1.0f);
		auxSolidCube(1.0f);

	glPopMatrix();

   // Draw the sphere next to the cube.
	glPushMatrix();

		glTranslatef(2.0f, 0.8f, -0.2f);
		glColor3f(1.0f, 1.0f ,1.0f);
      auxSolidSphere(0.7f);

	glPopMatrix();

   
	// Draw the object that is casting the shadow.
   glPushMatrix();

		glBindTexture(GL_TEXTURE_2D, ShadowObjectTex.ID);

		glBegin(GL_QUADS);

			glNormal3f(0, 1, 0);
         glTexCoord2f(0, 1);
			glVertex3f(ShadowObject[0].x, ShadowObject[0].y, ShadowObject[0].z);

         glTexCoord2f(0, 0);
			glVertex3f(ShadowObject[1].x, ShadowObject[1].y, ShadowObject[1].z);

         glTexCoord2f(1, 0);
			glVertex3f(ShadowObject[2].x, ShadowObject[2].y, ShadowObject[2].z);

         glTexCoord2f(1, 1);
			glVertex3f(ShadowObject[3].x, ShadowObject[3].y, ShadowObject[3].z);

		glEnd();

	glPopMatrix();


   // Last but not least we draw the light source.
	glDisable(GL_LIGHTING);
	glPushMatrix();

		glBindTexture(GL_TEXTURE_2D, NULL);
		
      // Here we are setting the light position.  We do this because the light position
      // can be moved with the arrow keys so we must update it.
		float lp[4];
		lp[0] = lightPosition.x; lp[1] = lightPosition.y;
		lp[2] = lightPosition.z; lp[3] = lightPosition.w;
		glLightfv(GL_LIGHT0, GL_POSITION, lp);

		glTranslatef(lightPosition.x, lightPosition.y, lightPosition.z);
		glColor3f(1.0f, 1.0f, 0.5f);
		auxSolidSphere(0.1);

	glPopMatrix();
	glEnable(GL_LIGHTING);
}


void DrawSilhouette(CVector4 lightPos, float Extend)
{
   CVector4 Extended;
   int A, B;

   // This function will take the silhouette of our object, create a shadow volume,
   // and draw the shadow volume (called after setting up the stencil buffer).
   glDisable(GL_LIGHTING);

   glColor3f(0.5f, 0.7f, 0.5f);

   // Loop through all the points to create the shadow volume.  Technically we would
   // normally want to calculate the silhouette but since this is a simple quad
   // it would not make much sense because all the points are...silhouette canidates.
   for(int i = 0; i < MAX_SHADOW_CASTER_POINTS; i++)
      {
          // These will hold which points we are working with.
          A = i; B = i + 1;

          // Error checking to make sure we don't go out of bounds.
          if(B >= MAX_SHADOW_CASTER_POINTS) B = 0;

         // Sine our object is a quad, we draw it as one.
         glBegin(GL_QUADS);

            // Pretty much we are creating a bunch quads coming off of the original.

            // First we start with the first point.
            glVertex3f(ShadowObject[A].x, ShadowObject[A].y, ShadowObject[A].z);

            // Next we extend that point based on the light position and the
            // objects position to form the second part of the invisible quad.
            // We are using a function of the Vector class called ExtendVertexPos
            // which will take a point, light position, and amount to extend and will
            // return the new position.
            Extended.ExtendVertexPos(ShadowObject[A], lightPos, Extend);
            glVertex3f(Extended.x, Extended.y, Extended.z);

            // Same as the top we take the second point and extend it to from
            // the third part of the quad.
            Extended.ExtendVertexPos(ShadowObject[B], lightPos, Extend);
            glVertex3f(Extended.x, Extended.y, Extended.z);

            // Finally we draw the second point as the last part of the quad.
            glVertex3f(ShadowObject[B].x, ShadowObject[B].y, ShadowObject[B].z);

         glEnd();
      }

   glEnable(GL_LIGHTING);
}


void Shutdown()
{
   // Free both textures that we loaded earlier.
   Terrain.FreeImage();
   ShadowObjectTex.FreeImage();
}


// Recap:

// Shadow volumes, as you can see when running this program, can add highly realistic
// shadows to any 3d scene.  There are many different techniques and I hope you find
// one that perfectly suites your needs.  Before we part until next time lets review
// what has happened in this demo.  First we set up the stencil buffer.  Next
// we draw only the shadow volumes to the buffer.  The shadow volumes are the silhouette
// edges of the object extended in a certain direction (direction is based on the light
// position).  After we draw the shadow volume for both lit and unlit areas we draw the
// scene twice for each matching area.  And we are done.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com